import pytest
import pickle
from copy import deepcopy
import sys
from collections.abc import MutableMapping
from .base import FrozendictTestBase


pyversion = sys.version_info
pyversion_major = pyversion[0]
pyversion_minor = pyversion[1]


class Map(MutableMapping):
    def __init__(self, *args, **kwargs):
        self._dict = dict(*args, **kwargs)
    
    def __getitem__(self, key):
        return self._dict[key]
    
    def __setitem__(self, key, value):
        self._dict[key] = value
    
    def __delitem__(self, key):
        del self._dict[key]
    
    def __iter__(self):
        return iter(self._dict)
    
    def __len__(self):
        return len(self._dict)


class FrozendictCommonTest(FrozendictTestBase):
    @property
    def is_mapping_implemented(self):
        return self.c_ext or (
            pyversion_major > 3 or 
            (pyversion_major == 3 and pyversion_minor > 9)
        )
    
    @property
    def is_reversed_implemented(self):
        return self.c_ext or (
            pyversion_major > 3 or 
            (pyversion_major == 3 and pyversion_minor > 7)
        )
    
    ####################################################################
    # main tests

    def test_bool_false(self, fd_empty):
        assert not fd_empty

    def test_constructor_kwargs(self, fd2, fd_dict_2):
        assert self.FrozendictClass(**fd_dict_2) == fd2

    def test_constructor_self(self, fd):
        assert fd == self.FrozendictClass(fd, Guzzanti="Corrado")

    def test_constructor_generator(self, fd, generator_seq2):
        assert fd == self.FrozendictClass(
            generator_seq2,
            Guzzanti="Corrado"
        )

    def test_constructor_hole(self, fd_hole, fd_dict_hole):
        assert fd_hole == self.FrozendictClass(fd_dict_hole)

    def test_constructor_map(self, fd_dict):
        assert self.FrozendictClass(Map(fd_dict)) == fd_dict

    def test_normalget(self, fd):
        assert fd["Guzzanti"] == "Corrado"

    def test_keyerror(self, fd):
        with pytest.raises(KeyError):
            fd["Brignano"]

    def test_len(self, fd, fd_dict):
        assert len(fd) == len(fd_dict)

    def test_in_true(self, fd):
        assert "Guzzanti" in fd

    def test_not_in_false(self, fd):
        assert not ("Guzzanti" not in fd)

    def test_in_false(self, fd):
        assert not ("Brignano" in fd)

    def test_not_in_true(self, fd):
        assert "Brignano" not in fd

    def test_bool_true(self, fd):
        assert fd

    def test_deepcopy_unhashable(self, fd_unhashable):
        fd_copy = deepcopy(fd_unhashable)
        assert fd_copy == fd_unhashable
        assert fd_copy is not fd_unhashable

    def test_not_equal(self, fd, fd2, fd_sabina):
        assert fd != fd_sabina
        assert fd != fd2
        assert not (fd == fd_sabina)
        assert not (fd == fd2)

    def test_equals_dict(self, fd, fd_dict):
        assert fd == fd_dict

    @pytest.mark.parametrize(
            "protocol",
            range(pickle.HIGHEST_PROTOCOL + 1)
    )
    def test_pickle(self, fd, protocol):
        dump = pickle.dumps(fd, protocol=protocol)
        assert dump
        assert pickle.loads(dump) == fd

    def test_constructor_iterator(self, fd, fd_items):
        assert self.FrozendictClass(fd_items) == fd

    def test_todict(self, fd, fd_dict):
        assert dict(fd) == fd_dict

    def test_get(self, fd):
        assert fd.get("Guzzanti") == "Corrado"

    def test_get_fail(self, fd):
        default = object()
        assert fd.get("Brignano", default) is default

    def test_keys(self, fd, fd_dict):
        assert tuple(fd.keys()) == tuple(fd_dict.keys())

    def test_values(self, fd, fd_dict):
        assert tuple(fd.values()) == tuple(fd_dict.values())

    def test_items(self, fd, fd_dict):
        assert tuple(fd.items()) == tuple(fd_dict.items())

    def test_fromkeys(self, fd_sabina):
        keys = ["Corrado", "Sabina"]
        f = self.FrozendictClass.fromkeys(keys, "Guzzanti")
        assert f == fd_sabina

    def test_fromkeys_dict(self, fd_sabina, fd_dict_sabina):
        f = self.FrozendictClass.fromkeys(fd_dict_sabina, "Guzzanti")
        assert f == fd_sabina

    def test_fromkeys_set(self, fd_sabina, fd_dict_sabina):
        f = self.FrozendictClass.fromkeys(
            set(fd_dict_sabina),
            "Guzzanti"
        )
        
        assert f == fd_sabina

    def test_repr(self, fd, fd_dict, module_prefix):
        classname = self.FrozendictClass.__name__
        dict_repr = repr(fd_dict)
        assert repr(fd) == f"{module_prefix}{classname}({dict_repr})"

    def test_str(self, fd, fd_dict, module_prefix):
        classname = self.FrozendictClass.__name__
        assert str(fd) == f"{module_prefix}{classname}({repr(fd_dict)})"

    def test_format(self, fd, fd_dict, module_prefix):
        classname = self.FrozendictClass.__name__
        dict_repr = repr(fd_dict)
        assert format(fd) == f"{module_prefix}{classname}({dict_repr})"

    def test_iter(self, fd):
        items = []

        for x in iter(fd):
            items.append((x, fd[x]))

        assert tuple(items) == tuple(fd.items())

    def test_sum(self, fd, fd_dict, fd_dict_sabina):
        new_fd = fd | fd_dict_sabina
        assert type(new_fd) is self.FrozendictClass
        new_dict = dict(fd_dict)
        new_dict.update(fd_dict_sabina)
        assert new_fd == new_dict

    def test_union(self, fd_dict, fd_sabina):
        new_fd = self.FrozendictClass(fd_dict)
        id_fd = id(new_fd)
        new_fd |= fd_sabina
        assert type(new_fd) is self.FrozendictClass
        assert id_fd != id(new_fd)
        new_dict = dict(fd_dict)
        new_dict.update(fd_sabina)
        assert new_fd == new_dict

    def test_reversed(self, fd, fd_dict):
        assert (tuple(reversed(fd)) == tuple(reversed(tuple(fd_dict))))

    def test_iter_len(self, fd):
        assert iter(fd).__length_hint__() >= 0

    def test_keys_len(self, fd):
        assert len(fd.keys()) == len(fd)

    def test_items_len(self, fd):
        assert len(fd.items()) == len(fd)

    def test_values_len(self, fd):
        assert len(fd.values()) == len(fd)

    def test_equal_keys(self, fd, fd_dict):
        assert fd.keys() == fd_dict.keys()

    def test_equal_items(self, fd, fd_dict):
        assert fd.items() == fd_dict.items()

    def test_in_keys(self, fd):
        assert "Guzzanti" in fd.keys()

    def test_in_items(self, fd):
        assert ("Guzzanti", "Corrado") in fd.items()

    def test_disjoint_true_keys(self, fd, fd_sabina):
        assert fd.keys().isdisjoint(fd_sabina)

    def test_disjoint_true_items(self, fd, fd_sabina):
        assert fd.items().isdisjoint(fd_sabina.items())

    def test_disjoint_false_keys(self, fd):
        assert not fd.keys().isdisjoint(fd)

    def test_disjoint_false_items(self, fd):
        assert not fd.items().isdisjoint(fd.items())

    def test_sub_keys(self, fd, fd_dict, fd2, fd_dict_2):
        res = frozenset(fd_dict.keys()) - frozenset(fd_dict_2.keys())
        assert fd.keys() - fd2.keys() == res

    def test_sub_items(self, fd, fd_dict, fd2, fd_dict_2):
        res = frozenset(fd_dict.items()) - frozenset(fd_dict_2.items())
        assert fd.items() - fd2.items() == res

    def test_and_keys(self, fd, fd_dict, fd2, fd_dict_2):
        res = frozenset(fd_dict.keys()) & frozenset(fd_dict_2.keys())
        assert fd.keys() & fd2.keys() == res

    def test_and_items(self, fd, fd_dict, fd2, fd_dict_2):
        res = frozenset(fd_dict.items()) & frozenset(fd_dict_2.items())
        assert fd.items() & fd2.items() == res

    def test_or_keys(self, fd, fd_dict, fd2, fd_dict_2):
        res = frozenset(fd_dict.keys()) | frozenset(fd_dict_2.keys())
        assert fd.keys() | fd2.keys() == res

    def test_or_items(self, fd, fd_dict, fd2, fd_dict_2):
        res = frozenset(fd_dict.items()) | frozenset(fd_dict_2.items())
        assert fd.items() | fd2.items() == res

    def test_xor_keys(self, fd, fd_dict, fd2, fd_dict_2):
        res = frozenset(fd_dict.keys()) ^ frozenset(fd_dict_2.keys())
        assert fd.keys() ^ fd2.keys() == res

    def test_xor_items(self, fd, fd_dict, fd2, fd_dict_2):
        res = frozenset(fd_dict.items()) ^ frozenset(fd_dict_2.items())
        assert fd.items() ^ fd2.items() == res

    @pytest.mark.parametrize(
            "protocol",
            range(pickle.HIGHEST_PROTOCOL + 1)
    )
    def test_pickle_iter_key(self, fd, protocol):
        orig = iter(fd.keys())
        dump = pickle.dumps(orig, protocol=protocol)
        assert dump
        assert tuple(pickle.loads(dump)) == tuple(orig)

    @pytest.mark.parametrize(
            "protocol",
            range(pickle.HIGHEST_PROTOCOL + 1)
    )
    def test_pickle_iter_item(self, fd, protocol):
        orig = iter(fd.items())
        dump = pickle.dumps(orig, protocol=protocol)
        assert dump
        assert tuple(pickle.loads(dump)) == tuple(orig)

    @pytest.mark.parametrize(
            "protocol",
            range(pickle.HIGHEST_PROTOCOL + 1)
    )
    def test_pickle_iter_value(self, fd, protocol):
        orig = iter(fd.values())
        dump = pickle.dumps(orig, protocol=protocol)
        assert dump
        assert tuple(pickle.loads(dump)) == tuple(orig)

    def test_lt_key(self, fd, fd_hole):
        assert fd_hole.keys() < fd.keys()

    def test_gt_key(self, fd, fd_hole):
        assert fd.keys() > fd_hole.keys()

    def test_le_key(self, fd, fd_hole):
        assert fd_hole.keys() <= fd.keys()
        assert fd.keys() <= fd.keys()

    def test_ge_key(self, fd, fd_hole):
        assert fd.keys() >= fd_hole.keys()
        assert fd.keys() >= fd.keys()

    def test_lt_item(self, fd, fd_hole):
        assert fd_hole.items() < fd.items()

    def test_gt_item(self, fd, fd_hole):
        assert fd.items() > fd_hole.items()

    def test_le_item(self, fd, fd_hole):
        assert fd_hole.items() <= fd.items()
        assert fd.items() <= fd.items()

    def test_ge_item(self, fd, fd_hole):
        assert fd.items() >= fd_hole.items()
        assert fd.items() >= fd.items()
    
    def test_mapping_keys(self, fd):
        if not self.is_mapping_implemented:
            pytest.skip("mapping not implemented")
        
        assert fd.keys().mapping == fd
    
    def test_mapping_items(self, fd):
        if not self.is_mapping_implemented:
            pytest.skip("mapping not implemented")
        
        assert fd.items().mapping == fd
    
    def test_mapping_values(self, fd):
        if not self.is_mapping_implemented:
            pytest.skip("mapping not implemented")
        
        assert fd.values().mapping == fd

    def test_reversed_keys(self, fd, fd_dict):
        if not self.is_reversed_implemented:
            pytest.skip("reversed not implemented")
            
        fd_keys = tuple(reversed(fd.keys()))
        dict_keys = tuple(reversed(tuple(fd_dict.keys())))
        assert fd_keys == dict_keys

    def test_reversed_items(self, fd, fd_dict):
        if not self.is_reversed_implemented:
            pytest.skip("reversed not implemented")
            
        fd_items = tuple(reversed(fd.items()))
        dict_items = tuple(reversed(tuple(fd_dict.items())))
        assert fd_items == dict_items

    def test_reversed_values(self, fd, fd_dict):
        if not self.is_reversed_implemented:
            pytest.skip("reversed not implemented")
            
        fd_values = tuple(reversed(fd.values()))
        dict_values = tuple(reversed(tuple(fd_dict.values())))
        assert fd_values == dict_values

    ####################################################################
    # frozendict-only tests

    def test_hash(self, fd, fd_eq):
        assert hash(fd)
        assert hash(fd) == hash(fd_eq)

    def test_unhashable_value(self, fd_unhashable):
        with pytest.raises(TypeError):
            hash(fd_unhashable)

        # hash is cached
        with pytest.raises(TypeError):
            hash(fd_unhashable)

    def test_set_replace(self, fd_dict, generator_seq2):
        items = tuple(generator_seq2)
        d2 = dict(items)
        assert fd_dict != d2
        f2 = self.FrozendictClass(items)
        fd3 = f2.set("Guzzanti", "Corrado")
        assert fd3 == fd_dict

    def test_set_add(self, fd_dict):
        d2 = dict(fd_dict, a="b")
        assert fd_dict != d2
        f2 = self.FrozendictClass(fd_dict)
        fd3 = f2.set("a", "b")
        assert fd3 == d2

    def test_setdefault_notinsert(self, fd, fd_dict):
        assert fd.setdefault("Hicks") is fd

    def test_setdefault_insert_default(self, fd, fd_dict):
        fd_dict.setdefault("Allen")
        assert fd_dict == fd.setdefault("Allen")

    def test_setdefault_insert(self, fd, fd_dict):
        fd_dict.setdefault("Allen", "Woody")
        assert fd_dict == fd.setdefault("Allen", "Woody")

    def test_del(self, fd_dict):
        d2 = dict(fd_dict)
        d2["a"] = "b"
        f2 = self.FrozendictClass(d2)
        fd3 = f2.delete("a")
        assert fd3 == fd_dict

    def test_key(self, fd):
        assert fd.key() == fd.key(0) == "Guzzanti"
        assert fd.key(1) == fd.key(-2) == "Hicks"

    def test_key_out_of_range(self, fd):
        with pytest.raises(IndexError):
            fd.key(3)
        
        with pytest.raises(IndexError):
            fd.key(-4)

    def test_value(self, fd):
        assert fd.value() == fd.value(0) == "Corrado"
        assert fd.value(1) == fd.value(-2) == "Bill"

    def test_value_out_of_range(self, fd):
        with pytest.raises(IndexError):
            fd.value(3)
        
        with pytest.raises(IndexError):
            fd.value(-4)

    def test_item(self, fd):
        assert fd.item() == fd.item(0) == ("Guzzanti", "Corrado")
        assert fd.item(1) == fd.item(-2) == ("Hicks", "Bill")

    def test_item_out_of_range(self, fd):
        with pytest.raises(IndexError):
            fd.item(3)
        
        with pytest.raises(IndexError):
            fd.item(-4)

    ####################################################################
    # immutability tests

    def test_normalset(self, fd):
        with pytest.raises(TypeError):
            fd["Guzzanti"] = "Caterina"

    def test_del_error(self, fd):
        with pytest.raises(TypeError):
            del fd["Guzzanti"]

    def test_clear(self, fd):
        with pytest.raises(AttributeError):
            fd.clear()

    def test_pop(self, fd):
        with pytest.raises(AttributeError):
            fd.pop("Guzzanti")

    def test_popitem(self, fd):
        with pytest.raises(AttributeError):
            fd.popitem()

    def test_update(self, fd):
        with pytest.raises(AttributeError):
            fd.update({"Brignano": "Enrico"})

    def test_delattr(self, fd):
        with pytest.raises(AttributeError):
            del fd.items



from __future__ import annotations

from typing import TypeVar, overload, Optional, Union

try:
    from typing import Mapping, Iterable, Iterator, Tuple, Type
except ImportError:
    from collections.abc import Mapping, Iterable, Iterator
    Tuple = tuple
    Type = type

K = TypeVar("K")
V = TypeVar("V", covariant=True)
K2 = TypeVar("K2")
V2 = TypeVar("V2", covariant=True)
SelfT = TypeVar("SelfT", bound=frozendict[K, V])

class frozendict(Mapping[K, V]):
    @overload
    def __new__(cls: Type[SelfT]) -> SelfT: ...
    @overload
    def __new__(cls: Type[SelfT], **kwargs: V) -> frozendict[str, V]: ...
    @overload
    def __new__(cls: Type[SelfT], mapping: Mapping[K, V]) -> SelfT: ...
    @overload
    def __new__(cls: Type[SelfT], iterable: Iterable[Tuple[K, V]]) -> SelfT: ...
    
    def __getitem__(self: SelfT, key: K) -> V: ...
    def __len__(self: SelfT) -> int: ...
    def __iter__(self: SelfT) -> Iterator[K]: ...
    def __hash__(self: SelfT) -> int: ...
    def __reversed__(self: SelfT) -> Iterator[K]: ...
    def copy(self: SelfT) -> SelfT: ...
    def __copy__(self: SelfT) -> SelfT: ...
    def __deepcopy__(self: SelfT) -> SelfT: ...
    def delete(self: SelfT, key: K) -> SelfT: ...
    @overload
    def key(self: SelfT, index: int) -> K: ...
    @overload
    def key(self: SelfT) -> K: ...
    @overload
    def value(self: SelfT, index: int) -> V: ...
    @overload
    def value(self: SelfT) -> V: ...
    @overload
    def item(self: SelfT, index: int) -> Tuple[K, V]: ...
    @overload
    def item(self: SelfT) -> Tuple[K, V]: ...
    @overload
    def __or__(self: SelfT, other: Mapping[K, V]) -> SelfT: ...
    @overload
    def __or__(self: SelfT, other: Mapping[K2, V]) -> frozendict[Union[K, K2], V]: ...
    @overload
    def __or__(self: SelfT, other: Mapping[K, V2]) -> frozendict[K, Union[V, V2]]: ...
    @overload
    def __or__(self: SelfT, other: Mapping[K2, V2]) -> frozendict[Union[K, K2], Union[V, V2]]: ...
    @overload
    def set(self: SelfT, key: K, value: V) -> SelfT: ...
    @overload
    def set(self: SelfT, key: K2, value: V) -> frozendict[Union[K, K2], V]: ...
    @overload
    def set(self: SelfT, key: K, value: V2) -> frozendict[K, Union[V, V2]]: ...
    @overload
    def set(self: SelfT, key: K2, value: V2) -> frozendict[Union[K, K2], Union[V, V2]]: ...
    @overload
    def setdefault(self: SelfT, key: K) -> SelfT: ...
    @overload
    def setdefault(self: SelfT, key: K2) -> SelfT: ...
    @overload
    def setdefault(self: SelfT, key: K, default: V) -> SelfT: ...
    @overload
    def setdefault(self: SelfT, key: K2, default: V) -> frozendict[Union[K, K2], V]: ...
    @overload
    def setdefault(self: SelfT, key: K, default: V2) -> frozendict[K, Union[V, V2]]: ...
    @overload
    def setdefault(self: SelfT, key: K2, default: V2) -> frozendict[Union[K, K2], Union[V, V2]]: ...
    
    @classmethod
    def fromkeys(
        cls: Type[SelfT], 
        seq: Iterable[K], 
        value: Optional[V] = None
    ) -> SelfT: ...


FrozenOrderedDict = frozendict
c_ext: bool

class FreezeError(Exception):  pass


class FreezeWarning(UserWarning):  pass



from .common import FrozendictCommonTest
from .subclass_only import FrozendictSubclassOnlyTest
import frozendict as cool
from frozendict import frozendict as FrozendictClass


is_subclass = True


class FrozendictSubclass(FrozendictClass):
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)


class FrozendictMissingSubclass(FrozendictClass):
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)
    
    def __missing__(self, key):
        return key


class TestFrozendictSubclass(
        FrozendictCommonTest, 
        FrozendictSubclassOnlyTest
):
    FrozendictClass = FrozendictSubclass
    FrozendictMissingClass = FrozendictMissingSubclass
    c_ext = cool.c_ext
    is_subclass = is_subclass



import pytest


class FrozendictTestBase:
    _FrozendictClass = None
    
    @property
    def FrozendictClass(self):
        val = self._FrozendictClass
        
        if val is None:
            raise ValueError("FrozendictClass is None")
        
        return val
    
    @FrozendictClass.setter
    def FrozendictClass(self, val):
        self._FrozendictClass = val
    
    _c_ext = None
    
    @property
    def c_ext(self):
        val = self._c_ext
        
        if val is None:
            raise ValueError("c_ext is None")
        
        return val
    
    @c_ext.setter
    def c_ext(self, val):
        self._c_ext = val
    
    _is_subclass = None
    
    @property
    def is_subclass(self):
        val = self._is_subclass
        
        if val is None:
            raise ValueError("is_subclass is None")
        
        return val
    
    @is_subclass.setter
    def is_subclass(self, val):
        self._is_subclass = val
    
    ####################################################################
    # dict fixtures

    @pytest.fixture
    def fd_dict(self):
        return {
            "Guzzanti": "Corrado", 
            "Hicks": "Bill", 
            self.FrozendictClass({1: 2}): "frozen"
        }

    @pytest.fixture
    def fd_dict_hole(self, fd_dict):
        new_dict = fd_dict.copy()
        del new_dict["Guzzanti"]
        return new_dict

    @pytest.fixture
    def fd_dict_eq(self):
        return {
            "Hicks": "Bill", 
            "Guzzanti": "Corrado", 
            self.FrozendictClass({1: 2}): "frozen"
        }

    @pytest.fixture
    def fd_dict_2(self):
        return {
            "Guzzanti": "Corrado", 
            "Hicks": "Bill", 
            "frozen": self.FrozendictClass({1: 2})
        }

    @pytest.fixture
    def fd_dict_sabina(self):
        return {'Corrado': 'Guzzanti', 'Sabina': 'Guzzanti'}

    @pytest.fixture
    def generator_seq2(self, fd_dict):
        seq2 = list(fd_dict.items())
        seq2.append(("Guzzanti", "Mario"))
        return (x for x in seq2)

    ####################################################################
    # frozendict fixtures

    @pytest.fixture
    def fd(self, fd_dict):
        return self.FrozendictClass(fd_dict)

    @pytest.fixture
    def fd_hole(self, fd_dict_hole):
        return self.FrozendictClass(fd_dict_hole)

    @pytest.fixture
    def fd_unhashable(self):
        return self.FrozendictClass({1: []})

    @pytest.fixture
    def fd_eq(self, fd_dict_eq):
        return self.FrozendictClass(fd_dict_eq)
    
    @pytest.fixture
    def fd2(self, fd_dict_2):
        return self.FrozendictClass(fd_dict_2)

    @pytest.fixture
    def fd_sabina(self, fd_dict_sabina):
        return self.FrozendictClass(fd_dict_sabina)

    @pytest.fixture
    def fd_items(self, fd_dict):
        return tuple(fd_dict.items())

    @pytest.fixture
    def fd_empty(self):
        return self.FrozendictClass()

    @pytest.fixture
    def module_prefix(self):
        if self.is_subclass:
            return ""
        
        return "frozendict."
    



_OldJsonEncoder = None
_oldOrjsonDumps = None
_oldMutableMappingSubclasshook = None


class MonkeypatchWarning(UserWarning):
    pass


def checkCExtension(*, warn, warn_c = False):
    import frozendict as cool
    
    res = cool.c_ext
    
    if warn and res == warn_c:
        if warn_c:
            msg = "C Extension version, monkeypatch will be not applied"
        else:
            msg = "Pure Python version, monkeypatch will be not applied"
        
        import warnings
        
        warnings.warn(msg, MonkeypatchWarning)
    
    return res


def patchOrUnpatchJson(*, patch, warn = True):
    if not checkCExtension(warn = warn):
        return
    
    from importlib import import_module
    self = import_module(__name__)
    import frozendict as cool
    import json
    
    OldJsonEncoder = self._OldJsonEncoder
    
    FrozendictJsonEncoder = cool._getFrozendictJsonEncoder(
        OldJsonEncoder
    )
    
    if patch:
        DefaultJsonEncoder = FrozendictJsonEncoder
    else:
        DefaultJsonEncoder = OldJsonEncoder
    
    if DefaultJsonEncoder is None:
        default_json_encoder = None
    else:
        default_json_encoder = DefaultJsonEncoder(
            skipkeys = False,
            ensure_ascii = True,
            check_circular = True,
            allow_nan = True,
            indent = None,
            separators = None,
            default = None,
        )
    
    if patch:
        if OldJsonEncoder is None:
            self._OldJsonEncoder = json.encoder.JSONEncoder
    else:
        if OldJsonEncoder is None:
            raise ValueError(
                "Old json encoder is None " +  
                "(maybe you already unpatched json?)"
            )
        
        self._OldJsonEncoder = None
        
    cool.FrozendictJsonEncoder = FrozendictJsonEncoder
    
    json.JSONEncoder = DefaultJsonEncoder
    json.encoder.JSONEncoder = DefaultJsonEncoder
    json._default_encoder = default_json_encoder
    

def patchOrUnpatchOrjson(*, patch, warn = True):
    if not checkCExtension(warn = warn):
        return
    
    from importlib import import_module
    self = import_module(__name__)
    import orjson
    
    if self._oldOrjsonDumps is None:
        if not patch:
            raise ValueError(
                "Old orjson encoder is None " + 
                "(maybe you already unpatched orjson?)"
            )
        
        oldOrjsonDumps = orjson.dumps
    else:
        oldOrjsonDumps = self._oldOrjsonDumps
    
    if patch:
        from frozendict import frozendict
        
        def frozendictOrjsonDumps(obj, *args, **kwargs):
            if isinstance(obj, frozendict):
                obj = dict(obj)
            
            return oldOrjsonDumps(obj, *args, **kwargs)
        
        defaultOrjsonDumps = frozendictOrjsonDumps
        newOldOrjsonDumps = oldOrjsonDumps
    else:
        defaultOrjsonDumps = oldOrjsonDumps
        newOldOrjsonDumps = None
    
    self._oldOrjsonDumps = newOldOrjsonDumps
    orjson.dumps = defaultOrjsonDumps
    orjson.orjson.dumps = defaultOrjsonDumps


def patchOrUnpatchMutableMappingSubclasshook(*, patch, warn = True):
    warn_c = True
    
    if checkCExtension(warn = warn, warn_c = warn_c):
        return
    
    from importlib import import_module
    self = import_module(__name__)
    from collections.abc import MutableMapping
    from frozendict import frozendict
    
    if self._oldMutableMappingSubclasshook is None:
        if not patch:
            raise ValueError(
                "Old MutableMapping subclasshook is None " + 
                "(maybe you already unpatched MutableMapping?)"
            )
        
        oldMutableMappingSubclasshook = MutableMapping.__subclasshook__
    else:
        oldMutableMappingSubclasshook = self._oldMutableMappingSubclasshook
    
    if patch:
        @classmethod
        def frozendictMutableMappingSubclasshook(
            klass, 
            subclass, 
            *args, 
            **kwargs
        ):
            if klass == MutableMapping:
                if issubclass(subclass, frozendict):
                    return False
                
                return oldMutableMappingSubclasshook(
                    subclass, 
                    *args, 
                    **kwargs
                )
            
            return NotImplemented
        
        defaultMutableMappingSubclasshook = frozendictMutableMappingSubclasshook
        newOldMutableMappingSubclasshook = oldMutableMappingSubclasshook
    else:
        defaultMutableMappingSubclasshook = oldMutableMappingSubclasshook
        newOldMutableMappingSubclasshook = None
    
    self._oldMutableMappingSubclasshook = newOldMutableMappingSubclasshook
    MutableMapping.__subclasshook__ = defaultMutableMappingSubclasshook
    
    try:
        MutableMapping._abc_caches_clear()
    except AttributeError:
        MutableMapping._abc_cache.discard(frozendict)
        MutableMapping._abc_negative_cache.discard(frozendict)


def patchOrUnpatchAll(*, patch, warn = True, raise_orjson = False):
    patchOrUnpatchJson(patch = patch, warn = warn)

    try:
        import orjson
    except ImportError:
        if raise_orjson:
            raise
    else:
        patchOrUnpatchOrjson(patch = patch, warn = warn)

    patchOrUnpatchMutableMappingSubclasshook(patch = patch, warn = warn)


__all__ = (
    patchOrUnpatchJson.__name__, 
    patchOrUnpatchOrjson.__name__, 
    patchOrUnpatchMutableMappingSubclasshook.__name__, 
    patchOrUnpatchAll.__name__, 
    MonkeypatchWarning.__name__, 
)



import pytest
import frozendict as cool
from frozendict import frozendict
from collections import OrderedDict
from collections.abc import MutableSequence, Sequence
from array import array
from types import MappingProxyType
from frozendict import FreezeError, FreezeWarning


class A:
    def __init__(self, x):
        self.x = x


class BaseSeq(Sequence):

    def __init__(self, seq):
        self._items = list(seq)
        
    def __getitem__(self, i):
        return self._items[i]
    
    def __len__(self):
        return len(self._items)    


class FrozenSeqA(BaseSeq):
    pass


class FrozenSeqB(BaseSeq):
    pass


class MutableSeq(BaseSeq, MutableSequence):
    
    def __delitem__(self, i):
        del self._items[i]
    
    def __setitem__(self, i, v):
        self._items[i] = v
        
    def insert(self, index, value):
        self._items.insert(index, value)


def custom_a_converter(a):
    return frozendict(**a.__dict__, y="mbuti")


def custom_inverse_converter(o):
    return dict(**o, y="mbuti")


@pytest.fixture
def before_cure_inverse():
    return [frozendict(a={1: 2})]


@pytest.fixture
def after_cure_inverse():
    return (frozendict(a=frozendict({1: 2}), y="mbuti"), )


@pytest.fixture
def a():
    a = A(3)
    
    return a


@pytest.fixture
def before_cure(a):
    return {"x": [
        5, 
        frozendict(y = {5, "b", memoryview(b"b")}), 
        array("B", (0, 1, 2)), 
        OrderedDict(a=bytearray(b"a")),
        MappingProxyType({2: []}),
        a
    ]}
    

@pytest.fixture
def after_cure():
    return frozendict(x = (
        5, 
        frozendict(y = frozenset({5, "b", memoryview(b"b")})), 
        (0, 1, 2), 
        frozendict(a = b'a'),
        MappingProxyType({2: ()}),
        frozendict(x = 3),
    ))


@pytest.fixture
def after_cure_a(a):
    return custom_a_converter(a)


def test_deepfreeze(before_cure, after_cure):
    assert cool.deepfreeze(before_cure) == after_cure


def test_register_bad_to_convert():
    with pytest.raises(ValueError):
        cool.register(5, 7)


def test_register_bad_converter():
    with pytest.raises(ValueError):
        cool.register(frozendict, 7)


def test_unregister_not_present():
    with pytest.raises(FreezeError):
        cool.unregister(frozendict)


def test_deepfreeze_bad_custom_converters_key(before_cure):
    with pytest.raises(ValueError):
        cool.deepfreeze(before_cure, custom_converters={7: 7})


def test_deepfreeze_bad_custom_converters_val(before_cure):
    with pytest.raises(ValueError):
        cool.deepfreeze(before_cure, custom_converters={frozendict: 7})


def test_deepfreeze_bad_custom_inverse_converters_key(before_cure):
    with pytest.raises(ValueError):
        cool.deepfreeze(before_cure, custom_inverse_converters={7: 7})


def test_deepfreeze_bad_custom_inverse_converters_val(before_cure):
    with pytest.raises(ValueError):
        cool.deepfreeze(
            before_cure,
            custom_inverse_converters = {frozendict: 7}
        )


def test_register_custom(a):
    cool.register(A, custom_a_converter)
    assert cool.deepfreeze(a) == custom_a_converter(a)
    cool.unregister(A)
    assert cool.deepfreeze(a) == frozendict(a.__dict__)
    
    
def test_deepfreeze_custom(a):
    assert cool.deepfreeze(
        a,
        custom_converters={A: custom_a_converter}
    ) == custom_a_converter(a)


def test_deepfreeze_inverse(before_cure_inverse, after_cure_inverse):
    assert cool.deepfreeze(
        before_cure_inverse,
        custom_inverse_converters={frozendict: custom_inverse_converter}
    ) == after_cure_inverse


def test_register_inverse(before_cure_inverse, after_cure_inverse):
    with pytest.warns(FreezeWarning):
        cool.register(
            frozendict,
            custom_inverse_converter,
            inverse=True
        )
    
    assert cool.deepfreeze(before_cure_inverse) == after_cure_inverse


def test_prefer_forward():
    assert isinstance(
        cool.deepfreeze(
            [FrozenSeqA((0, 1, 2))],
            custom_converters={FrozenSeqA: FrozenSeqB},
            custom_inverse_converters={FrozenSeqA: MutableSeq}
            )[0],
        FrozenSeqB
        )

def test_original_immutate():
    unfrozen = {
        "int": 1,
        "nested": {"int": 1},
    }

    frozen = cool.deepfreeze(unfrozen)
    
    assert type(unfrozen["nested"]) is dict
